from typing import *
import os
from subprocess import call
import tempfile
import shutil

import pypdf
from pdfminer.pdfparser import PDFParser
from pdfminer.pdfdocument import PDFDocument, PDFEncryptionError
from pdfminer.pdfinterp import PDFResourceManager, PDFPageInterpreter
from pdfminer.converter import PDFPageAggregator
from pdfminer.layout import LAParams, LTTextBox, LTTextLine, LTChar, LTFigure, LTPage, Rect
from pdfminer.pdfpage import PDFPage
import pytesseract

from pdf_reader.custom_dataclasses import PdfReaderConfig, NaturalTextHelper
from pdf_reader.pdf_page import ParseePdfPage
from pdf_reader.helper import make_images_from_pdf


"""
Currently the image detection is not very sophisticated and goes purely by file extensions
This is in order not to add some additional packages like python-magic just for this simple check
"""
def is_image(file_path: str):
    image_types_supported = ["png", "jpg", "jpeg"]
    file_name = os.path.basename(file_path).lower()
    if "." in file_name:
        ending = file_name.rsplit(".", 1)
        if ending[-1] in image_types_supported:
            return True
    return False


def decrypt_pdf_with_qpdf(path):
    # make temp file path
    repaired_file_path = path[0:-4] + "_repaired.pdf"

    # convert
    call('qpdf --password=%s --decrypt %s %s' % ('', path, repaired_file_path), shell=True)

    # delete old file
    os.remove(path)

    # rename repaired file
    os.rename(repaired_file_path, path)


def open_pdf(pdf_path: str) -> Tuple[PDFDocument, PDFPageInterpreter, PDFPageAggregator, any, pypdf.PdfReader]:
    # Open a PDF file.
    fp = open(pdf_path, 'rb')
    # Create a PDF parser object associated with the file object.
    parser = PDFParser(fp)
    # try to open document normally
    try:
        document = PDFDocument(parser, password='')
    except PDFEncryptionError:
        fp.close()
        # try to convert doc with qpdf
        decrypt_pdf_with_qpdf(pdf_path)
        # open converted file
        fp = open(pdf_path, 'rb')
        # Create a PDF parser object associated with the file object.
        parser = PDFParser(fp)
        # try to open document again
        document = PDFDocument(parser, password='')
    # Create a PDF resource manager object that stores shared resources.
    rsrcmgr = PDFResourceManager()
    # Create a PDF device object.
    laparams = LAParams()
    device = PDFPageAggregator(rsrcmgr, laparams=laparams)
    # Create a PDF interpreter object.
    interpreter = PDFPageInterpreter(rsrcmgr, device)
    # pypdf
    pypdf_reader = pypdf.PdfReader(pdf_path)
    return document, interpreter, device, fp, pypdf_reader


def get_natural_text(reader: pypdf.PdfReader, page_index: int) -> NaturalTextHelper:
    try:
        pypdf_text = reader.pages[page_index].extract_text()
    except Exception:
        pypdf_text = None
    return NaturalTextHelper(pypdf_text)


def get_pdf_pages(pdf_path: str, config: Optional[PdfReaderConfig] = None, force_ocr: bool = False, **kwargs) -> List[ParseePdfPage]:
    config = PdfReaderConfig(None, None, None) if config is None else config
    # check if file is an image
    if is_image(pdf_path):
        mediabox, text_boxes, pypdf_reader = get_elements_from_image(pdf_path)
        return [ParseePdfPage(0, pdf_path, mediabox, text_boxes, config, get_natural_text(pypdf_reader, 0))]
    document, interpreter, device, fp, pypdf_reader = open_pdf(pdf_path)
    pages = []
    for page_index, page in enumerate(PDFPage.create_pages(document)):
        interpreter.process_page(page)
        layout = device.get_result()
        text_boxes = parse_layout(layout)
        run_ocr = force_ocr or needs_ocr(text_boxes)
        if run_ocr:
            mediabox, text_boxes, pypdf_reader = repair_layout(pdf_path, page_index)
            page_obj = ParseePdfPage(page_index, pdf_path, mediabox, text_boxes, config, get_natural_text(pypdf_reader, page_index))
        else:
            page_obj = ParseePdfPage(page_index, pdf_path, page.mediabox, text_boxes, config, get_natural_text(pypdf_reader, page_index))
        pages.append(page_obj)

    fp.close()
    return pages


def parse_layout(layout_obj: any, force_chars: bool = False) -> List[Union[LTTextBox, LTChar]]:
    all_relevant_elements = []
    """Function to recursively parse the layout tree and flatten LTFigure elements."""
    for lt_obj in layout_obj:
        if isinstance(lt_obj, LTFigure):
            all_relevant_elements += parse_layout(lt_obj)  # Recursive
        elif isinstance(lt_obj, LTTextBox):
            if force_chars:
                all_relevant_elements += parse_layout(lt_obj._objs)
            else:
                all_relevant_elements.append(lt_obj)
        elif isinstance(lt_obj, LTChar):
            all_relevant_elements.append(lt_obj)
        elif isinstance(lt_obj, LTTextLine):
            all_relevant_elements += parse_layout(lt_obj._objs)
    return all_relevant_elements


# page needs OCR if either no elements found or unreadable characters are present
def needs_ocr(text_boxes: List[Union[LTTextBox, LTChar]]) -> bool:

    if len(text_boxes) == 0:
        return True

    for element in text_boxes:
        if isinstance(element, LTTextBox):
            for o in element._objs:
                if isinstance(o, LTTextLine):
                    try:
                        text = o.get_text()
                    except Exception as e:
                        text = ""
                    text_stripped = text.strip()
                    if text_stripped:
                        for kk, c in enumerate(o._objs):
                            t = c.get_text()
                            if t is not None and t.startswith("(cid:"):
                                return True
    return False


def get_elements_from_image(image_path: str, custom_temp_folder_path: Optional[str] = None) -> Tuple[any, List[LTTextBox], pypdf.PdfReader]:

    if custom_temp_folder_path is None:
        temp_folder = tempfile.TemporaryDirectory()
        temp_folder_path = temp_folder.name
    else:
        temp_folder_path = custom_temp_folder_path

    # start tesseract
    pdf_path_tmp = os.path.join(temp_folder_path, "ocr.pdf")
    f = open(pdf_path_tmp, "w+b")
    pdf = pytesseract.image_to_pdf_or_hocr(image_path, extension='pdf')
    f.write(bytearray(pdf))
    f.close()

    document, interpreter, device, fp, pypdf_reader = open_pdf(pdf_path_tmp)
    pages = [x for x in PDFPage.create_pages(document)]
    if len(pages) != 1:
        raise Exception("repaired PDF page does not have exactly one page")

    page = pages[0]
    interpreter.process_page(page)
    layout = device.get_result()
    # delete temporary folder and contents
    if custom_temp_folder_path is None:
        shutil.rmtree(temp_folder_path)
    # use char level parsing for image output
    boxes = parse_layout(layout, True)
    fp.close()
    return page.mediabox, boxes, pypdf_reader


def repair_layout(path: str, page_index: int) -> Tuple[any, List[LTTextBox], pypdf.PdfReader]:

    temp_folder = tempfile.TemporaryDirectory()
    temp_folder_path = temp_folder.name

    # save temporary image of PDF
    target_height = 3200
    img_path_tmp = make_images_from_pdf(path, temp_folder_path, [target_height], page_index)[target_height][0]

    output = get_elements_from_image(img_path_tmp, temp_folder_path)
    shutil.rmtree(temp_folder_path)
    return output



import copy
import json
from typing import *
import re
from dataclasses import dataclass
from functools import reduce

from pdf_reader.helper import clean_numeric_value


class PdfReaderConfig:
    SPACE_MAX_DISTANCE = 6
    TOLERANCE_GEN = 10
    LINE_BREAK_MAX_DISTANCE = 14

    # default page size, all coordinates will be transformed to fit these default max widths
    page_default_width_normal = 594
    page_default_width_horizontal = 1200

    tolerance_top_bot = 3
    tolerance_top_bot_line_items = 6
    char_dist_max = 1
    line_break_distance = 20
    separate_columns_distance = 150
    separate_columns_distance_wide_layout = 250
    separate_table_distance = 50
    text_max_number_col = 15
    table_header_max_height = 120
    table_sides_tolerance_x0 = 50
    table_sides_tolerance_x1 = 50
    tolerance_columns_li = 20
    line_max_height = 2

    def __init__(self, space_max_distance: Union[int, None], tolerance_gen: Union[int, None], line_break_distance: Union[int, None]):
        if space_max_distance is not None:
            self.SPACE_MAX_DISTANCE = space_max_distance

        if tolerance_gen is not None:
            self.TOLERANCE_GEN = tolerance_gen

        if line_break_distance is not None:
            self.LINE_BREAK_MAX_DISTANCE = line_break_distance


@dataclass
class RelativeAreaPrediction:
    class_name: str
    x0: float
    x1: float
    y0: float
    y1: float
    prob: float


class Rectangle:

    def __init__(self, x0: Union[int, None], x1: Union[int, None], y0: Union[int, None], y1: Union[int, None]):

        self.x0 = x0
        self.x1 = x1
        self.y0 = y0
        self.y1 = y1

        self.x0_el = None
        self.x1_el = None
        self.y0_el = None
        self.y1_el = None

        self.in_group = False

        self.tolerance_detection = 3

    def __str__(self):
        return "R: (x0: " + str(self.x0) + ", x1: " + str(self.x1) + ", y0: " + str(self.y0) + ", y1: " + str(
            self.y1) + ")"

    def __repr__(self):
        return "R: (x0: " + str(self.x0) + ", x1: " + str(self.x1) + ", y0: " + str(self.y0) + ", y1: " + str(
            self.y1)

    def list(self):
        return [round(self.x0), round(self.x1), round(self.y0), round(self.y1)]

    def dict_coordinates(self):
        return {"x0": self.x0, "x1": self.x1, "y0": self.y0, "y1": self.y1}

    def width(self):
        return self.x1 - self.x0

    def height(self):
        return self.y1 - self.y0

    def x_distance_to(self, element):

        if self.h_overlap(element, "", 0):
            return 0

        if element.x0 >= self.x1:
            return element.x0 - self.x1
        else:
            return abs(self.x0 - element.x1)

    def y_distance_to(self, element):

        if self.v_overlap(element, "", 0):
            return 0

        if element.y0 >= self.y1:
            return element.y0 - self.y1
        else:
            return abs(self.y0 - element.y1)

    def v_overlap(self, element, field_add="", custom_tolerance=None):

        tolerance = custom_tolerance if custom_tolerance is not None else self.tolerance_detection

        # make element a bit smaller for detection
        y0_chosen = getattr(self, "y0" + field_add) + tolerance
        y1_chosen = getattr(self, "y1" + field_add) - tolerance

        v_overlaps = True

        if (y1_chosen < element.y0) or (y0_chosen > element.y1):
            v_overlaps = False
        return v_overlaps

    def list_index(self, area_list):
        for a, area in enumerate(area_list):
            if area.x0 == self.x0 and area.x1 == self.x1 and area.y0 == self.y0 and area.y1 == self.y1:
                return a
        return -1

    def h_inside(self, element, field_add="", custom_tolerance=None):

        tolerance = custom_tolerance if custom_tolerance is not None else self.tolerance_detection

        # make element a bit larger for detection
        x0_chosen = getattr(self, "x0" + field_add) - tolerance
        x1_chosen = getattr(self, "x1" + field_add) + tolerance

        el_x0_chosen = getattr(element, "x0" + field_add)
        el_x1_chosen = getattr(element, "x1" + field_add)

        if el_x0_chosen >= x0_chosen and el_x1_chosen <= x1_chosen:
            return True
        else:
            return False

    def h_overlap(self, element, field_add="", custom_tolerance=None):

        tolerance = custom_tolerance if custom_tolerance is not None else self.tolerance_detection

        # make element a bit smaller for detection
        x0_chosen = getattr(self, "x0" + field_add) + tolerance
        x1_chosen = getattr(self, "x1" + field_add) - tolerance

        h_overlaps = True

        if (x0_chosen > element.x1) or (x1_chosen < element.x0):
            h_overlaps = False
        return h_overlaps

    def overlap_percent(self, area):

        h_percent = self.h_overlap_percent([area])[0]
        v_percent = self.v_overlap_percent([area])[0]

        return (h_percent + v_percent) / 2

    def h_overlap_percent(self, area_list, field_add=""):

        x0_chosen = getattr(self, "x0" + field_add)
        x1_chosen = getattr(self, "x1" + field_add)

        overlaps = []
        for area in area_list:
            if not self.h_overlap(area, field_add, 0):
                overlaps.append(0)
                continue
            if x0_chosen <= area.x0 and x1_chosen >= area.x1:
                overlaps.append(area.width() / self.width())
                continue
            if area.x0 <= x0_chosen and area.x1 >= x1_chosen:
                overlaps.append(1)
                continue
            if x0_chosen <= area.x0:
                overlaps.append(min(1, (x1_chosen - area.x0) / self.width()))
            else:
                # x0_chosen > area.x0
                overlaps.append(min(1, (area.x1 - x0_chosen) / self.width()))

        return overlaps

    def v_overlap_percent(self, area_list, field_add=""):

        y0_chosen = getattr(self, "y0" + field_add)
        y1_chosen = getattr(self, "y1" + field_add)

        overlaps = []
        for area in area_list:
            if not self.v_overlap(area, field_add, 0):
                overlaps.append(0)
                continue
            if y0_chosen <= area.y0 and y1_chosen >= area.y1:
                overlaps.append(area.height() / self.height())
                continue
            if area.y0 <= y0_chosen and area.y1 >= y1_chosen:
                overlaps.append(1)
                continue
            if y0_chosen <= area.y0:
                overlaps.append(min(1, (y1_chosen - area.y0) / self.height()))
            else:
                # y0_chosen > area.y0
                overlaps.append(min(1, (area.y1 - y0_chosen) / self.height()))

        return overlaps

    def collides_with(self, element, field_add=""):

        return self.h_overlap(element, field_add) and self.v_overlap(element, field_add)

    def is_inside(self, element, custom_tolerance=None):

        tolerance = custom_tolerance if custom_tolerance is not None else self.tolerance_detection

        # make element a bit larger for detection
        x0_chosen = self.x0 - tolerance
        x1_chosen = self.x1 + tolerance
        y0_chosen = self.y0 - tolerance
        y1_chosen = self.y1 + tolerance

        if element.x0 >= x0_chosen and element.x1 <= x1_chosen and element.y0 >= y0_chosen and element.y1 <= y1_chosen:
            return True
        else:
            return False


class AreaPrediction(Rectangle):

    def __init__(self, relative_area: RelativeAreaPrediction, page_width: int, page_height: int, class_id: int):
        super().__init__(int(relative_area.x0 * page_width), int(relative_area.x1 * page_width), int((1 - relative_area.y1) * page_height), int((1 - relative_area.y0) * page_height))
        self.class_value = relative_area.class_name
        self.class_id = class_id
        self.prob = relative_area.prob


class SingleChar(Rectangle):

    def __init__(self, x0, x1, y0, y1, text, scale_multiplier=1):
        super().__init__(x0 * scale_multiplier, x1 * scale_multiplier, y0 * scale_multiplier, y1 * scale_multiplier)

        self.text = text

    def __str__(self):
        return str(self.text) + " [" + str(round(self.x0, 2)) + "," + str(round(self.x1, 2)) + "," + str(
            round(self.y0, 2)) + "," + str(round(self.y1, 2)) + "]"

    def __repr__(self):
        return str(self)


def make_area_from_elements(elements: List[Rectangle], field_add: str = "") -> Rectangle:
    min_x0 = None
    min_y0 = None
    max_x1 = None
    max_y1 = None

    for el in elements:
        elx0 = getattr(el, "x0" + field_add)
        elx1 = getattr(el, "x1" + field_add)
        ely0 = getattr(el, "y0" + field_add)
        ely1 = getattr(el, "y1" + field_add)

        min_x0 = elx0 if min_x0 is None or elx0 < min_x0 else min_x0
        min_y0 = ely0 if min_y0 is None or ely0 < min_y0 else min_y0
        max_x1 = elx1 if max_x1 is None or elx1 > max_x1 else max_x1
        max_y1 = ely1 if max_y1 is None or ely1 > max_y1 else max_y1

    return Rectangle(min_x0, max_x1, min_y0, max_y1)


class BaseElement(Rectangle):

    def __init__(self, x0: int = 0, x1: int = 0, y0: int = 0, y1: int = 0, text: str = "", row_index: Optional[int] = None, has_bold: bool = False,
                 scale_multiplier: float = 1):

        super().__init__(int(x0 * scale_multiplier), int(x1 * scale_multiplier), int(y0 * scale_multiplier), int(y1 * scale_multiplier))

        self.text = text
        self.has_bold = has_bold

        self.row_index = row_index

        self.tolerance_detection = 1

        self.in_group = False
        self.linked_el = []

    def __str__(self):
        b_val = "True" if self.has_bold else "False"
        return "t:'" + str(self.text) + "', x0: " + str(round(self.x0)) + ", x1: " + str(
            round(self.x1)) + ", y0: " + str(round(self.y0)) + ", y1: " + str(round(self.y1)) + ", b: " + b_val + ""

    def __repr__(self):
        return str(self.text) + " (" + str(round(self.x0)) + "," + str(round(self.x1)) + "," + str(
            round(self.y0)) + "," + str(round(self.y1)) + ")"

    # convert data to dict so that it can be saved as e.g. JSON
    def to_dict(self):
        return {"type": "em", "a": self.list(), "t": self.text}

    def is_identical(self, element):

        if round(element.x0) == round(self.x0) and round(element.x1) == round(self.x1) and round(element.y0) == round(
                self.y0) and round(element.y1) == round(self.y1) and self.text == element.text:
            return True
        else:
            return False

    def in_list(self, el_list):
        for el in el_list:
            if self.is_identical(el):
                return True
        return False

    def collides_with(self, element, field_add=""):

        # make element a bit smaller for detection
        x0_chosen = self.x0 + self.tolerance_detection
        x1_chosen = self.x1 - self.tolerance_detection
        y0_chosen = self.y0 + self.tolerance_detection
        y1_chosen = self.y1 - self.tolerance_detection

        hoverlaps = True
        voverlaps = True
        if (x0_chosen > element.x1) or (x1_chosen < element.x0):
            hoverlaps = False
        if (y1_chosen < element.y0) or (y0_chosen > element.y1):
            voverlaps = False

        return hoverlaps and voverlaps

    def merge(self, element):

        # change text
        # check y position difference
        y_diff = abs(element.y1 - self.y1)
        if y_diff <= 2:
            # same y position
            if self.x0 < element.x0:
                self.text = self.text + " " + element.text
            else:
                self.text = element.text + " " + self.text
        else:
            if element.y1 > self.y1:
                self.text = element.text + " " + self.text
            else:
                self.text = self.text + " " + element.text

        # change coordinates
        self.x0 = element.x0 if element.x0 < self.x0 else self.x0
        self.x1 = element.x1 if element.x1 > self.x1 else self.x1
        self.y0 = element.y0 if element.y0 < self.y0 else self.y0
        self.y1 = element.y1 if element.y1 > self.y1 else self.y1

        self.row_index = min(self.row_index, element.row_index)


def natural_text_elements(elements: List[BaseElement]) -> str:
    elements_sorted = sorted(elements, key=lambda x: (x.row_index, x.x0))
    return " ".join([x.text for x in elements_sorted])


class BaseElementGroup(BaseElement):
    elements: List[BaseElement]

    def __init__(self, elements: List[BaseElement]):
        super().__init__()
        self.elements = elements
        self.fit_elements()

    def fit_elements(self, field_add=""):
        new_area = make_area_from_elements(self.elements, field_add)
        self.x0 = new_area.x0
        self.x1 = new_area.x1
        self.y0 = new_area.y0
        self.y1 = new_area.y1
        self.text = natural_text_elements(self.elements)
        self.has_bold = False
        for el in self.elements:
            if el.has_bold:
                self.has_bold = True
                break
        self.row_index = min([x.row_index for x in self.elements]) if len(self.elements) > 0 else 0

    def add_element(self, element: BaseElement):
        self.elements.append(element)
        self.fit_elements()


class Area(Rectangle):
    elements: List[BaseElement]

    def __init__(self, x0=0, x1=0, y0=0, y1=0, tolerance_detection=3, elements: Optional[List[BaseElement]] = None):

        super().__init__(x0, x1, y0, y1)

        self.elements = [] if elements is None else elements

        self.in_group = False

        self.tolerance_detection = tolerance_detection

        self.all_row_indices = set()

    def __str__(self):
        el_str = ", ".join(["(" + str(x) + ")" for x in self.elements])
        return "A: (x0: " + str(self.x0) + ", x1: " + str(self.x1) + ", y0: " + str(self.y0) + ", y1: " + str(
            self.y1) + ", elements: [" + el_str + "])"

    def __repr__(self):
        el_str = ", ".join(["(" + str(x) + ")" for x in self.elements])
        return "A: (x0: " + str(self.x0) + ", x1: " + str(self.x1) + ", y0: " + str(self.y0) + ", y1: " + str(
            self.y1) + ", elements: [" + el_str + "])"

    def break_at_horizontal(self, break_point, discard=None):

        if break_point >= self.y1:
            return self, None
        if break_point <= self.y0:
            return self, None

        former_y0 = self.y0

        if discard == "top":
            self.y1 = break_point
            return self, None
        elif discard == "bot":
            self.y0 = break_point
            return self, None
        else:
            self.y0 = break_point
            new_area = Area(self.x0, self.x1, former_y0, break_point)
            # split elements
            to_del = []
            for k, el in enumerate(self.elements):
                if not self.is_inside(el):
                    new_area.elements.append(el)
                    to_del.append(k)
            for k in range(len(self.elements) - 1, -1, -1):
                if k in to_del:
                    del (self.elements[k])

            # fit width of areas again
            self.refit_x_only()
            new_area.refit_x_only()

            return self, new_area

    def put_element(self, el):

        self.elements.append(el)
        self.fit_elements(el)

    def init_with_elements(self, el_list):

        self.elements = [x for x in el_list if x is not None]

        self.x0 = None
        self.x1 = None
        self.y0 = None
        self.y1 = None

        for el in self.elements:
            if self.x0 is None or el.x0 < self.x0:
                self.x0 = el.x0
            if self.y0 is None or el.y0 < self.y0:
                self.y0 = el.y0
            if self.x1 is None or el.x1 > self.x1:
                self.x1 = el.x1
            if self.y1 is None or el.y1 > self.y1:
                self.y1 = el.y1

    def fit_elements(self, el_fit=None):

        self.all_row_indices = set()
        if el_fit is None:
            el_list = self.elements
        else:
            el_list = [el_fit]

        for el in el_list:
            if self.x0_el is None or el.x0 < self.x0_el:
                self.x0_el = el.x0
            if self.y0_el is None or el.y0 < self.y0_el:
                self.y0_el = el.y0
            if self.x1_el is None or el.x1 > self.x1_el:
                self.x1_el = el.x1
            if self.y1_el is None or el.y1 > self.y1_el:
                self.y1_el = el.y1
            self.all_row_indices.add(el.row_index)

    def refit_x_only(self):

        if len(self.elements) == 0:
            return

        self.x0 = None
        self.x1 = None
        self.x0_el = None
        self.x1_el = None

        for el in self.elements:
            if self.x0_el is None or el.x0 < self.x0_el:
                self.x0_el = el.x0
                self.x0 = el.x0
            if self.x1_el is None or el.x1 > self.x1_el:
                self.x1_el = el.x1
                self.x1 = el.x1

    def stackable(self, el, field_add=""):

        if self.h_inside(el, field_add) or el.h_inside(self, field_add) or self.h_overlap_percent([el], field_add)[
            0] > 0.6:
            # check that elements don't collide
            joined_areas_a = self.get_joined_areas()
            joined_areas_b = el.get_joined_areas()

            for ar in joined_areas_a:
                for ar2 in joined_areas_b:
                    if ar.collides_with(ar2):
                        return False
            return True
        else:
            return False

    def get_joined_areas(self):

        els_sorted = sorted(self.elements, key=lambda x: -x.y1)

        separate_areas = []
        el_list_temp = []
        for k in range(0, len(els_sorted)):
            if k == 0:
                el_list_temp.append(els_sorted[k])
            else:
                if els_sorted[k - 1].y0 - els_sorted[k].y1 < self.tolerance_detection:
                    el_list_temp.append(els_sorted[k])
                else:
                    # create new area with previous elements
                    new_area = Area(0, 0, 0, 0)
                    new_area.init_with_elements(el_list_temp)
                    separate_areas.append(new_area)
                    el_list_temp = [els_sorted[k]]
            if k == len(els_sorted) - 1 and len(el_list_temp) > 0:
                # create new area with previous elements
                new_area = Area(0, 0, 0, 0)
                new_area.init_with_elements(el_list_temp)
                separate_areas.append(new_area)
        return separate_areas

    def combine(self, el):

        self_copy = copy.deepcopy(self)

        if el.x0 < self.x0:
            self_copy.x0 = el.x0
        if el.y0 < self.y0:
            self_copy.y0 = el.y0
        if el.x0_el is not None and self.x0_el is not None:
            if el.x0_el < self.x0_el:
                self_copy.x0_el = el.x0_el
        if el.y0_el is not None and self.y0_el is not None:
            if el.y0_el < self.y0_el:
                self_copy.y0_el = el.y0_el

        if el.x1 > self.x1:
            self_copy.x1 = el.x1
        if el.y1 > self.y1:
            self_copy.y1 = el.y1
        if el.x1_el is not None and self.x1_el is not None:
            if el.x1_el > self.x1_el:
                self_copy.x1_el = el.x1_el
        if el.y1_el is not None and self.y1_el is not None:
            if el.y1_el > self.y1_el:
                self_copy.y1_el = el.y1_el

        for el_contained in el.elements:
            self_copy.elements.append(el_contained)

        self_copy.fit_elements()

        return self_copy

    def contains(self, element: BaseElement) -> bool:
        for el in self.elements:
            if el.is_identical(element):
                return True
            if isinstance(el, BaseElementGroup) and not isinstance(element, BaseElementGroup):
                for el_contained in el.elements:
                    if el_contained.is_identical(element):
                        return True
        return False


class TableGroup(Rectangle):
    elements: List[Area]

    def __init__(self, elements: List[Area], field_add=""):

        super().__init__(None, None, None, None)

        self.elements = [] + elements
        self.fit_elements(field_add)

        self.data_area = None
        self.line_items = []

    def __str__(self):
        el_str = ", ".join(["(" + str(x) + ")" for x in self.elements])
        return "G: (x0: " + str(self.x0) + ", x1: " + str(self.x1) + ", y0: " + str(self.y0) + ", y1: " + str(
            self.y1) + ", elements: [" + el_str + "])"

    def __repr__(self):
        el_str = ", ".join(["(" + str(x) + ")" for x in self.elements])
        return "G: (x0: " + str(self.x0) + ", x1: " + str(self.x1) + ", y0: " + str(self.y0) + ", y1: " + str(
            self.y1) + ", elements: [" + el_str + "])"

    def put_element(self, element, field_add=""):

        element.in_group = True
        self.elements.append(element)
        self.fit_elements(field_add)

    def remove_element(self, index: int, field_add=""):

        self.elements[index].in_group = False
        del (self.elements[index])
        self.fit_elements(field_add)

    def fit_elements(self, field_add=""):
        new_area = make_area_from_elements(self.elements, field_add)
        self.x0 = new_area.x0
        self.x1 = new_area.x1
        self.y0 = new_area.y0
        self.y1 = new_area.y1

    def elements_by_row(self) -> Dict[int, List[BaseElement]]:
        output = {}
        for area in self.elements:
            for el in area.elements:
                if el.row_index not in output:
                    output[el.row_index] = []
                output[el.row_index].append(el)
        return output


class ValueItem:

    def __init__(self, el):
        self.el = el
        self.val = el.text if el is not None else ""
        self.val_clean = None
        self.make_final_value()

    def __str__(self):
        return str(self.val)

    def __repr__(self):
        return self.__str__()

    def is_empty(self):
        if self.val.strip() == "":
            return True
        return False

    def make_final_value(self):

        if self.val == "":
            return

        self.val_clean = clean_numeric_value(self.val)

    def dict(self, simple=False):

        if simple:
            return {"v": self.val_clean}

        return {
            "v": self.val,
            "n": self.val_clean
        }


class LineItem:
    caption: str
    values: List[ValueItem]
    base_elements: Set[BaseElement]
    el: BaseElementGroup

    def __init__(self, el: Union[BaseElement, BaseElementGroup], num_cols: Optional[int] = None):
        self.base_elements = set()
        self.el = BaseElementGroup([])
        self.add_el(el)
        self.values = [ValueItem(None) for _ in range(0, num_cols)] if num_cols is not None else []

    def __str__(self):
        return "LI " + str(self.caption) + "; values: " + str(self.values)

    def __repr__(self):
        return self.__str__()

    def has_line_item_values(self):
        return False in [x.is_empty() for x in self.values]

    # adds element to line item
    def add_el(self, el: Union[BaseElement, BaseElementGroup]):
        if not isinstance(el, BaseElementGroup):
            self.el.add_element(el)
            self.base_elements.add(el)
        else:
            for contained in el.elements:
                self.el.add_element(contained)
                self.base_elements.add(contained)
        self.caption = self.el.text

    def add_value(self, el: BaseElement, col_idx: int):
        if col_idx <= len(self.values) - 1 and self.values[col_idx].is_empty():
            self.values[col_idx] = ValueItem(el)

    def assign_values(self, value_el_list):
        if len(self.values) != 0:
            raise Exception("should be called only with empty values")
        for k, v_el in enumerate(value_el_list):
            self.values.append(ValueItem(v_el))

    def first_non_empty_value(self):
        for v in self.values:
            if not v.is_empty():
                return v
        return None

    def dict(self, simple_values=False):
        return {
            "c": self.caption,
            "a": self.el.list(),
            "v": [v.dict(simple_values) for v in self.values]
        }


class NaturalTextHelper:
    text_raw: Union[str, None]
    lines: List[str]
    lines_cleaned: List[str]

    def __init__(self, pypdf_text: Optional[str]):
        self.text_raw = pypdf_text
        if pypdf_text is not None:
            self.lines = pypdf_text.split("\n")
            self.lines_cleaned = [self.clean_text_for_matching(x) for x in self.lines]
        else:
            self.lines = []
            self.lines_cleaned = []

    def clean_text_for_matching(self, string_val: str) -> str:
        return re.sub(r'[^A-Za-z.,\d]', '', string_val).lower()

    def has_line_items(self, group: TableGroup) -> bool:
        for row_idx, items in group.elements_by_row().items():
            items_sorted = sorted(items, key=lambda x: -x.x1)
            if len(items_sorted) == 1:
                last_item_text = self.clean_text_for_matching(items_sorted[0].text)
            else:
                last_item_text = self.clean_text_for_matching(items_sorted[1].text + items_sorted[0].text)
            for line in self.lines_cleaned:
                if line.endswith(last_item_text):
                    return True
        return False

    def is_adjacent_percent(self, table: TableGroup, compare_area: Area) -> float:
        all_rows = table.elements_by_row()
        # go row by row
        matches = 0
        for row_idx, base_elements in all_rows.items():
            relevant_elements = sorted([x for x in compare_area.elements if x.row_index == base_elements[0].row_index], key=lambda x: x.x0)
            if len(relevant_elements) > 0:
                if len(relevant_elements) == 1:
                    text_li = self.clean_text_for_matching(relevant_elements[0].text)
                else:
                    text_li = self.clean_text_for_matching(relevant_elements[0].text + relevant_elements[1].text)

                base_el_sorted = sorted(base_elements, key=lambda x: x.x0)
                if len(base_elements) == 1:
                    item_text = self.clean_text_for_matching(base_el_sorted[0].text)
                else:
                    item_text = self.clean_text_for_matching(base_el_sorted[0].text + base_el_sorted[1].text)

                for line in self.lines_cleaned:
                    if line.startswith(text_li) and item_text in line:
                        matches += 1
                        break
        return matches / len(all_rows.keys())


class ExtractedPdfElement(Rectangle):
    el: Optional[BaseElement]
    in_area: Optional[AreaPrediction]

    def __init__(self, x0: int, x1: int, y0: int, y1: int, el: Optional[BaseElement]):
        super().__init__(x0, x1, y0, y1)
        self.el = el

    def __str__(self):
        return str(self.el.to_dict() if self.el is not None else None)

    def __repr__(self):
        return self.__str__()

    def get_text(self) -> str:
        return self.el.text if self.el is not None else ""


class ExtractedFigure(ExtractedPdfElement):

    def __init__(self, x0: int, x1: int, y0: int, y1: int):
        super().__init__(x0, x1, y0, y1, None)

    def get_text(self) -> str:
        return "[figure]"


class PdfParagraph(ExtractedPdfElement):

    def __init__(self, elements: List[ExtractedPdfElement], config: PdfReaderConfig):
        super().__init__(0, 0, 0, 0, None)
        self.config = config
        self.text = ""
        self.line_break_char = "\n"
        self.elements: List[ExtractedPdfElement] = elements
        self.fit_area_from_elements()
        self.add_text_from_elements()
        self.area_group_ids = []

    def __str__(self):
        return self.text

    def __repr__(self):
        return str(self)

    def get_text(self) -> str:
        return self.text

    def to_dict(self):
        return {"text": self.text, "coordinates": self.dict_coordinates()}

    def toJSON(self):
        return json.dumps(self.to_dict())

    def add_text_from_elements(self):
        self.text = ""
        elements_sorted = sorted(self.elements, key=lambda el: (-el.y1, el.x0))
        for k, el in enumerate(elements_sorted):
            if k > 0 and self.elements[k - 1].y1 - el.y1 > self.config.LINE_BREAK_MAX_DISTANCE:
                self.text += self.line_break_char
            elif k > 0:
                self.text += " "
            self.text += el.get_text()

    def fit_area_from_elements(self):
        area = make_area_from_elements(self.elements)
        self.x0 = area.x0
        self.x1 = area.x1
        self.y0 = area.y0
        self.y1 = area.y1
        self.area_group_ids = set([el.in_area.class_id for el in self.elements if el.in_area is not None])

    def add_el(self, el: ExtractedPdfElement):
        self.elements.append(el)
        self.fit_area_from_elements()
        self.add_text_from_elements()


class ExtractedTable(ExtractedPdfElement):
    table_area: Area
    total_value_area: Area
    value_areas: List[Area]
    line_item_area: Area
    items: List[LineItem]
    g_index: int
    _items_by_row_idx: Dict[int, LineItem]
    num_rows: int
    num_cols: int
    _DEFAULT_SPACE_COLS = 5

    def __init__(self, line_items: List[LineItem], g_index: int):
        super().__init__(0, 0, 0, 0, None)
        self.g_index = g_index
        self.set_line_items(line_items)

    def fill_empty_li(self, value_grid: Dict[int, List[List[Union[BaseElement, None]]]]):

        final_li: List[LineItem] = []
        num_cols = 0
        for row_index, val_list in value_grid.items():
            num_cols = len(val_list) if len(val_list) > 0 else num_cols
            # find li
            chosen_li = None
            for li in self.items:
                if li.el.row_index == row_index:
                    chosen_li = li
                    break
            # create empty line item if not found
            if chosen_li is None:
                default_el = next((item for item in val_list if item is not None), None)
                if default_el is None:
                    raise Exception("empty values")
                else:
                    default_y0 = default_el.y0
                    default_y1 = default_el.y1
                chosen_li = LineItem(BaseElement(x0=self.line_item_area.x0, x1=self.line_item_area.x1, y0=default_y0, y1=default_y1, row_index=row_index))
            chosen_li.assign_values(val_list)
            final_li.append(chosen_li)

        self.set_line_items(final_li)

        # check that all items have the necessary columns
        for li in self.items:
            if len(li.values) == 0:
                li.assign_values([None for _ in range(0, num_cols)])
            elif len(li.values) != num_cols:
                raise Exception("number of columns not consistent")

    def set_line_items(self, line_items: List[LineItem]):

        self.items = list(sorted(line_items, key=lambda x: x.el.row_index))
        self.set_areas()
        self.set_table_size()
        self._items_by_row_idx = {}

        for li in line_items:
            self._items_by_row_idx[li.el.row_index] = li

        super().__init__(self.table_area.x0, self.table_area.x1, self.table_area.y0, self.table_area.y1, None)

    def __str__(self):
        return "T: (r: " + str(self.num_rows) + ", c:" + str(self.num_cols) + ")"

    def __repr__(self):
        return self.__str__()

    def to_dict(self):

        return {
            "type": "et",
            "i": [li.dict() for li in self.items],
            "a": self.table_area.list(),
            "va": [v.list() for v in self.value_areas],
            "tva": self.total_value_area.list(),
        }

    def df_format(self):

        table_data = []

        if len(table_data) > 0:
            table_data.append({"item": ""})

        for li in self.items:
            row_dict = {"item": li.caption}
            for k, col in enumerate(li.values):
                row_dict['col_' + str(k)] = col.val

            table_data.append(row_dict)

        return table_data

    def set_areas(self):
        self.line_item_area = Area()
        self.total_value_area = Area()
        self.table_area = Area()
        self.value_areas = []
        if len(self.items) == 0:
            return
        li_area_elements: List[BaseElement] = []
        value_area_elements: List[List[BaseElement]] = [[] for _ in self.items[0].values]
        for li in self.items:
            li_area_elements.append(li.el)
            for k, val in enumerate(li.values):
                value_area_elements[k].append(val.el)
        total_value_area_elements = reduce(lambda acc, x: acc + x, value_area_elements, [])
        all_elements = total_value_area_elements + li_area_elements
        self.line_item_area.init_with_elements(li_area_elements)
        self.total_value_area.init_with_elements(total_value_area_elements)
        self.table_area.init_with_elements(all_elements)
        for elements in value_area_elements:
            area = Area()
            area.init_with_elements(elements)
            self.value_areas.append(area)

        # adjust value areas to extend full space they have, all but first col
        value_areas = sorted([x for x in self.value_areas if len(x.elements) > 0], key=lambda x: x.x0)
        for a in range(len(value_areas) - 1, 0, -1):
            value_areas[a].x0 = min(value_areas[a].x0,
                                         value_areas[a - 1].x1 + self._DEFAULT_SPACE_COLS)
            value_areas[a].y1 = max(value_areas[a].y1, self.line_item_area.y1)
            value_areas[a].y0 = min(value_areas[a].y0, self.line_item_area.y0)
        # first column
        if len(value_areas) > 1:
            value_areas[0].x0 = min(value_areas[0].x0,
                                         value_areas[0].x1 - value_areas[1].width())
            value_areas[0].y1 = max(value_areas[0].y1, self.line_item_area.y1)
            value_areas[0].y0 = min(value_areas[0].y0, self.line_item_area.y0)
        # line items: close gap to first li
        if len(value_areas) > 0:
            self.line_item_area.x1 = max(self.line_item_area.x1, value_areas[0].x0 - self._DEFAULT_SPACE_COLS)

    def set_table_size(self):

        self.num_cols = 0
        self.num_rows = len(self.items)
        if self.num_rows > 0:
            self.num_cols = len(self.items[0].values)

    def add_value(self, element: BaseElement, col_idx: int):
        if element.row_index in self._items_by_row_idx:
            self._items_by_row_idx[element.row_index].add_value(element, col_idx)

    def remove_column(self, column_index: int):
        for li in self.items:
            del (li.values[column_index])
        self.set_line_items(self.items)

    def add_to_items(self, base_el: BaseElementGroup):
        if base_el.row_index not in self.line_item_area.all_row_indices:
            # add empty line item
            new_item = LineItem(base_el, len(self.items[0].values))
            line_items = copy.copy(self.items)
            line_items.append(new_item)
            self.set_line_items(line_items)
        else:
            # merge with best line item caption
            self._items_by_row_idx[base_el.row_index].add_el(base_el)
        self.set_areas()


@dataclass
class ExtractedPage:
    index: int
    size: Rectangle
    elements: List[ExtractedPdfElement]
    paragraphs: List[ExtractedPdfElement]



from typing import *

from pdf_reader.custom_dataclasses import Rectangle, ExtractedPdfElement
from pdf_reader.helper import make_images_from_pdf
from pdf_reader.pdf_page import ParseePdfPage


from PIL import Image, ImageDraw, ImageFont


class Displayer:

    def __init__(self, output_path: str, size: Rectangle, jpg_path: Optional[str]):

        self.path = output_path
        if jpg_path is None:
            self.im = Image.new('RGB', (round(size.width()), round(size.height())), (255, 255, 255))
        else:
            self.im = Image.open(jpg_path)
        self.draw = ImageDraw.Draw(self.im, 'RGBA')
        self.page_height = size.height()
        self.page_width = size.width()
        self.font = ImageFont.truetype("/Library/Fonts/Arial.ttf", 8)

    def draw_elements(self, element_list: List[Rectangle], color: Optional[Tuple]=None, outline: Optional[Tuple]=None):

        for k, box in enumerate(element_list):
            self.draw.rectangle([(box.x0, self.page_height - box.y1), (box.x1, self.page_height - box.y0)], fill=color,
                                outline=outline)

    def create_image(self):
        self.im.save(self.path, "PNG")


def visualise_elements(page: ParseePdfPage, elements: List[Rectangle], output_image_dir: str):

    color = (220, 0, 0, 50)

    # make jpg from page
    size = page.page_size.width() if page.has_wide_layout() else page.page_size.height()
    paths = make_images_from_pdf(page.pdf_path, output_image_dir, [size], page.page_index)
    jpg_path = paths[size][0]
    d = Displayer(jpg_path, page.page_size, jpg_path)
    d.draw_elements(elements, color, color)
    d.create_image()





import os
import re
import shutil
import tempfile
from decimal import Decimal
from subprocess import call
import cv2
from pdf2image import convert_from_path, pdfinfo_from_path
from typing import List, Dict, Optional

IMG_BATCH_SIZE = 10

to_filter_numbers = re.compile(r'(\([^0-9 ]*\))|[^0-9A-Za-z/]')


def letter_len(string):
    string = re.sub('[^A-Za-z]', '', string)
    return len(string)


def words_contained(cell_str, lower=False):
    if lower:
        cell_str = cell_str.lower()
    return list(filter(lambda x: x != "", re.sub('[^A-Za-z0-9%$€£¥]', ' ', cell_str).split(" ")))


def is_number_cell(cell_str):
    if cell_str is None:
        return False
    cell_str = to_filter_numbers.sub('', cell_str)
    if cell_str.isdigit():
        return True
    else:
        return False


def scale_image(path, target_max_width_or_height, target_width=None):
    image = cv2.imread(path)
    if target_width is not None:
        resized = cv2.resize(image, (round(target_width), round(target_max_width_or_height)),
                             interpolation=cv2.INTER_AREA)
    else:
        if image.shape[0] > image.shape[1]:
            ratio = float(image.shape[0]) / float(image.shape[1])
            target_width = int(target_max_width_or_height / ratio)
            resized = cv2.resize(image, (target_width, target_max_width_or_height), interpolation=cv2.INTER_AREA)
        else:
            ratio = float(image.shape[1]) / float(image.shape[0])
            target_height = int(target_max_width_or_height / ratio)
            resized = cv2.resize(image, (target_max_width_or_height, target_height), interpolation=cv2.INTER_AREA)

    cv2.imwrite(path, resized)


def get_target_size(width: int, height: int, target_size: int):

    if width > height:
        target_height = int((target_size / width) * height)
        target_width = target_size
    else:
        target_height = target_size
        target_width = int((target_size / height) * width)

    return target_width, target_height


def make_images_from_pdf(path_to_pdf: str, output_path: str, target_sizes: List[int], page_index_only: Optional[int]) -> Dict[int, List[str]]:

    def convert_batch(first_page_num: int, last_page_num: int):
        pages = convert_from_path(path_to_pdf, first_page=first_page_num, last_page=last_page_num, fmt="jpg")
        for page_index, image in enumerate(pages):
            for target_size in target_sizes:
                if target_size not in image_paths:
                    image_paths[target_size] = []
                image = image.resize(get_target_size(image.width, image.height, target_size))
                image_path = os.path.join(output_path, f"{target_size}_p_{page_index + first_page_num -1}.jpg")
                image.save(image_path, 'JPEG')
                image_paths[target_size].append(image_path)

    info = pdfinfo_from_path(path_to_pdf)
    max_pages = info["Pages"]
    image_paths = {}
    if page_index_only is None:
        for page_batch in range(1, max_pages + 1, IMG_BATCH_SIZE):
            convert_batch(page_batch, min(page_batch + IMG_BATCH_SIZE - 1, max_pages))
    else:
        convert_batch(page_index_only+1, page_index_only+1)

    return image_paths


def comma_separator_thousands(cell_str):
    if re.search(r'\b[0-9]{1,3}[,][0-9]{3}\b', cell_str):
        return True
    return False


def dot_separator_thousands(cell_str):
    if re.search(r'\b[0-9]{1,3}[.][0-9]{3}\b', cell_str):
        return True
    return False


def is_negative(cell_str):
    # minus
    if re.search(r'(-|—|–|‒|―|–|−)( | |)*\d', cell_str.strip()):
        return True
    # brackets
    if re.search(r'\([\d ,.%]+(\)|\b)', cell_str.strip()):
        return True
    return False


def clean_numeric_value(cell_str):
    mult = 1
    if is_negative(cell_str):
        mult = -1

    cell_str = re.sub(r'[^0-9,.]', '', cell_str)

    # clean thousands separator
    if comma_separator_thousands(cell_str):
        cell_str = re.sub(r',', "", cell_str)
    elif dot_separator_thousands(cell_str):
        cell_str = re.sub(r'\.', "", cell_str)

    # now also replace the comma with a dot should it be used in any case
    cell_str = re.sub(r',', ".", cell_str)

    if cell_str.replace('.', '', 1).isdigit():
        return Decimal(cell_str) * mult
    else:
        return None



from typing import *
import math
import pypdf

from pdf_reader.custom_dataclasses import *
from pdf_reader.converter import get_pdf_pages, is_image
from pdf_reader.visualisation import visualise_elements


def relative_areas_to_area_predictions(relative_areas: List[RelativeAreaPrediction], page_width: int, page_height: int) -> List[AreaPrediction]:
    return sorted([AreaPrediction(x, page_width, page_height, k) for k, x in enumerate(relative_areas) if x.prob > 0.5], key=lambda x: -x.prob)


def make_paragraphs(elements: List[ExtractedPdfElement], config: PdfReaderConfig, element_areas: Union[None, List[AreaPrediction]], all_text: Union[None, str], page_width: int) -> List[
    ExtractedPdfElement]:
    elements = sorted(elements, key=lambda x: x.y1, reverse=True)
    current_groups: List[PdfParagraph] = []
    all_groups: List[ExtractedPdfElement] = []
    GROUP_TOLERANCE = 10
    element_areas = [] if element_areas is None else element_areas

    natural_text = all_text.replace("\n", " ") if all_text is not None else None

    figures = [x for x in element_areas if x.class_value == "figure"]

    for el in elements:
        # check if el is contained in some detected area
        el.in_area = None
        for area in element_areas:
            if area.is_inside(el):
                el.in_area = area
                break
        if el.in_area in figures:
            # handle figures separately
            continue
        if not isinstance(el, ExtractedTable):
            added_to_group = False
            # check if element can be added to a current group
            for n in range(len(current_groups) - 1, -1, -1):
                if current_groups[n].y0 - el.y1 > GROUP_TOLERANCE:
                    del current_groups[n]
                elif current_groups[n].y0 - el.y1 <= config.LINE_BREAK_MAX_DISTANCE:
                    # check if el can be added to group
                    add_to_group = False
                    if current_groups[n].h_inside(el) or current_groups[n].h_overlap(el):
                        add_to_group = True
                    elif current_groups[n].x_distance_to(el) <= config.SPACE_MAX_DISTANCE and el.in_area is not None and el.in_area.class_id in current_groups[n].area_group_ids:
                        add_to_group = True
                    elif natural_text is not None and current_groups[n].x_distance_to(el) <= config.SPACE_MAX_DISTANCE and (
                            ((current_groups[n].elements[-2].get_text() + " ") if len(current_groups[n].elements) > 1 else "") + current_groups[n].elements[
                        -1].get_text() + " " + el.get_text() in natural_text):
                        add_to_group = True
                    if add_to_group:
                        current_groups[n].add_el(el)
                        added_to_group = True
                        break
            # create new group if not added to existing one
            if not added_to_group:
                p = PdfParagraph([el], config)
                current_groups.append(p)
                all_groups.append(p)
        else:
            all_groups.append(el)

    for figure in figures:
        all_groups.append(ExtractedFigure(figure.x0, figure.x1, figure.y0, figure.y1))

    return list(sorted(all_groups, key=lambda el: (1 if el.x0 > (page_width / 2) else 0, -(math.floor(el.y1 / 40)))))


"""
Returns a list of elements (text, tables) from a PDF or image.
file_path can be path to PDF or image (currently supported: "png", "jpg", "jpeg")
"""
def get_elements_from_pdf(file_path: str, detected_areas: Union[None, Dict[int, List[RelativeAreaPrediction]]] = None, force_ocr: bool = False, **kwargs) -> List[ExtractedPage]:
    pages = get_pdf_pages(file_path, None, force_ocr, **kwargs)

    all_pages = []

    for page_index, p in enumerate(pages):
        page_elements = p.extract_text_and_tables(**kwargs)
        areas = None
        if detected_areas is not None and page_index in detected_areas:
            areas = relative_areas_to_area_predictions(detected_areas[page_index], p.page_size.width(), p.page_size.height())
        paragraphs = make_paragraphs(page_elements, PdfReaderConfig(20, 10, 6), areas, p.natural_text.text_raw, p.page_size.width())
        all_pages.append(ExtractedPage(page_index, p.page_size, page_elements, paragraphs))

    return all_pages


def visualise_pdf_output(pdf_path: str, output_path: str, force_ocr: bool = False, **kwargs):
    pages = get_pdf_pages(pdf_path, None, force_ocr, **kwargs)
    for page_index, p in enumerate(pages):
        page_elements = p.extract_text_and_tables(**kwargs)
        visualise_elements(p, page_elements, output_path)



from typing import *
import math
import copy

import numpy as np
from pdfminer.layout import LTTextBox, LTTextLine, LTChar, Rect

from pdf_reader.custom_dataclasses import Rectangle, BaseElement, Area, BaseElementGroup, LineItem, ValueItem, ExtractedTable, ExtractedPdfElement, PdfReaderConfig, TableGroup, NaturalTextHelper
from pdf_reader.helper import is_number_cell, letter_len, words_contained


def filter_out_empty_columns(output_list: List[ExtractedTable], min_cols: int) -> List[ExtractedTable]:
    # filter out columns which have only None values
    for table_index in range(len(output_list) - 1, -1, -1):
        table = output_list[table_index]
        if len(table.items) > 0:
            for col_index in range(len(table.items[0].values) - 1, -1, -1):
                if len([li.values[col_index] for li in table.items if
                        li.values[col_index].is_empty()]) == len(table.items):
                    # splice column
                    table.remove_column(col_index)
        if len(table.items) == 0 or len(table.items[0].values) < min_cols:
            del (output_list[table_index])

    return output_list


def has_collision_post_alignment(list1, list2, new_pos, vertical, take_min):
    list_a = copy.deepcopy(list1)
    list_b = copy.deepcopy(list2)

    if vertical:
        if take_min:
            key = "x0"
            key_o = "x1"
        else:
            key = "x1"
            key_o = "x0"
    else:
        if take_min:
            key = "y0"
            key_o = "y1"
        else:
            key = "y1"
            key_o = "y0"

    # align object
    for el in list_a:
        prior_val = getattr(el, key)
        diff = new_pos - prior_val
        if diff != 0:
            setattr(el, key, new_pos)
            setattr(el, key_o, getattr(el, key_o) + diff)
    for el in list_b:
        prior_val = getattr(el, key)
        diff = new_pos - prior_val
        if diff != 0:
            setattr(el, key, new_pos)
            setattr(el, key_o, getattr(el, key_o) + diff)

    for el in list_a:
        for el2 in list_b:
            if not el.is_identical(el2):
                if el.collides_with(el2):
                    return True
    return False


class ParseePdfPage:

    page_index: int
    page_size: Rectangle
    config: PdfReaderConfig
    scale_multiplier: float = 1
    elements_list: List[BaseElement]
    non_text_elements: List[BaseElement]
    natural_text: NaturalTextHelper

    def __init__(self, page_index: int, pdf_path: str, page_size_pdfminer: Rect, text_boxes: List[Union[LTTextBox, LTChar]], config: PdfReaderConfig, natural_text: NaturalTextHelper):

        self.page_index = page_index
        self.pdf_path = pdf_path
        self.config = config
        self._set_page_size(page_size_pdfminer)
        self._set_elements(text_boxes)
        self.natural_text = natural_text

    def _get_page_size_multiplier(self, mediabox: Rect):

        page_width = mediabox[2] - mediabox[0]
        page_height = mediabox[3] - mediabox[1]

        if page_height > page_width:
            scale_multiplier = float(self.config.page_default_width_normal / page_width)
        else:
            scale_multiplier = float(self.config.page_default_width_horizontal / page_width)

        return scale_multiplier

    def has_wide_layout(self):
        return self.page_size.width() > self.page_size.height()

    def _set_page_size(self, page_size_pdfminer: Rect):
        self.scale_multiplier = self._get_page_size_multiplier(page_size_pdfminer)
        self.page_size = Rectangle(int(page_size_pdfminer[0] * self.scale_multiplier), int(page_size_pdfminer[2] * self.scale_multiplier), int(page_size_pdfminer[1] * self.scale_multiplier), int(page_size_pdfminer[3] * self.scale_multiplier))

    def _set_elements(self, text_boxes: List[Union[LTTextBox, LTChar]]):

        self.elements_list = []
        self.non_text_elements = []

        # now collect elements
        new_text = ""
        new_x0 = None
        new_x1 = None
        new_y0 = None
        new_y1 = None
        has_bold = False

        ref = None
        last_char_ref = None
        break_now = False
        for kk, element in enumerate(text_boxes):
            if isinstance(element, LTTextBox):
                for o in element._objs:
                    if isinstance(o, LTTextLine):
                        try:
                            text = o.get_text()
                        except Exception as e:
                            text = ""
                        text_stripped = text.strip()
                        if text_stripped:

                            # treat all space separated text fragments as separate objects, or if there is a harsh break in size/y-position
                            new_text = ""
                            new_x0 = None
                            new_x1 = None
                            new_y0 = None
                            new_y1 = None
                            has_bold = False

                            ref_y0 = None
                            ref_y1 = None
                            ref_size = None
                            break_now = False
                            for kk, c in enumerate(o._objs):

                                t = c.get_text()
                                if isinstance(c, LTChar) and t != " " and t != "\t" and t != ")":
                                    if ref_y0 is None:
                                        ref_y0 = c.y0
                                        ref_y1 = c.y1
                                        ref_size = c.size
                                    elif abs(ref_y0 - c.y0) > 1 or abs(ref_y1 - c.y1) > 1 or abs(ref_size - c.size) > 3:
                                        break_now = True

                                    if not break_now:
                                        new_x0 = c.x0 if new_x0 is None or new_x0 > c.x0 else new_x0
                                        new_x1 = c.x1 if new_x1 is None or new_x1 < c.x1 else new_x1
                                        new_y0 = c.y0 if new_y0 is None else new_y0
                                        new_y1 = c.y1 if new_y1 is None else new_y1
                                        new_text += t
                                        font_name = c.fontname.lower()
                                        if "bold" in font_name:
                                            has_bold = True

                                # start new element
                                if t == " " or t == "\t" or t == ")" or break_now:
                                    if t == ")":
                                        new_text += ")"
                                        new_x1 = c.x1
                                    if new_text != "" and new_x0 is not None and new_x1 is not None:
                                        self.elements_list.append(
                                            BaseElement(x0=new_x0, x1=new_x1, y0=new_y0, y1=new_y1, text=new_text,
                                                        has_bold=has_bold, scale_multiplier=self.scale_multiplier))
                                    has_bold = False
                                    if not break_now:
                                        new_text = ""
                                        new_x0 = None
                                        new_x1 = None
                                        new_y0 = None
                                        new_y1 = None
                                    else:
                                        new_text = t
                                        new_x0 = c.x0
                                        new_x1 = c.x1
                                        new_y0 = c.y0
                                        new_y1 = c.y1
                                        font_name = c.fontname.lower()
                                        if "bold" in font_name:
                                            has_bold = True

                                    ref_y0 = None
                                    ref_y1 = None
                                    ref_size = None
                                    break_now = False
                                if kk == len(
                                        o._objs) - 1 and new_text != "" and new_x0 is not None and new_x1 is not None:
                                    self.elements_list.append(
                                        BaseElement(x0=new_x0, x1=new_x1, y0=new_y0, y1=new_y1, text=new_text,
                                                    has_bold=has_bold, scale_multiplier=self.scale_multiplier))
                                    # reset variables
                                    new_text = ""
                                    new_x0 = None
                                    new_x1 = None
                                    new_y0 = None
                                    new_y1 = None

            elif isinstance(element, LTChar):
                # process char elements directly
                # treat all space separated text fragments as separate objects, or if there is a harsh break in size/y-position
                c = element
                t = c.get_text()
                if t != " " and t != "\t" and t != ")":
                    if ref is None:
                        ref = c
                    elif abs(ref.y0 - c.y0) > 1 or abs(ref.y1 - c.y1) > 1 or abs(ref.size - c.size) > 3:
                        break_now = True

                    # break if y distance too far
                    if kk > 0 and abs(c.y1 - text_boxes[kk - 1].y1) > 1:
                        break_now = True

                    # break if x distance too far
                    if kk > 0 and (abs(c.x0 - text_boxes[kk - 1].x1) > self.config.char_dist_max or (
                            last_char_ref is not None and abs(c.x0 - last_char_ref.x1) > self.config.char_dist_max)):
                        break_now = True

                    if not break_now:
                        new_x0 = c.x0 if new_x0 is None or new_x0 > c.x0 else new_x0
                        new_x1 = c.x1 if new_x1 is None or new_x1 < c.x1 else new_x1
                        new_y0 = c.y0 if new_y0 is None else new_y0
                        new_y1 = c.y1 if new_y1 is None else new_y1
                        new_text += t
                        font_name = c.fontname.lower()
                        if "bold" in font_name:
                            has_bold = True

                    last_char_ref = c

                # start new element
                if (t == " " or t == "\t" or t == ")" or break_now):
                    if t == ")":
                        new_text += ")"
                        new_x1 = c.x1
                    if new_text != "" and new_x0 is not None and new_x1 is not None:
                        self.elements_list.append(
                            BaseElement(x0=new_x0, x1=new_x1, y0=new_y0, y1=new_y1, text=new_text, has_bold=has_bold,
                                        scale_multiplier=self.scale_multiplier))
                    has_bold = False
                    if not break_now:
                        new_text = ""
                        new_x0 = None
                        new_x1 = None
                        new_y0 = None
                        new_y1 = None
                    else:
                        new_text = t
                        new_x0 = c.x0
                        new_x1 = c.x1
                        new_y0 = c.y0
                        new_y1 = c.y1
                        font_name = c.fontname.lower()
                        if "bold" in font_name:
                            has_bold = True

                    ref = None
                    break_now = False
                if kk == len(text_boxes) - 1 and new_text != "" and new_x0 is not None and new_x1 is not None:
                    self.elements_list.append(
                        BaseElement(x0=new_x0, x1=new_x1, y0=new_y0, y1=new_y1, text=new_text, has_bold=has_bold,
                                    scale_multiplier=self.scale_multiplier))
                    # no need to reset variables as end is reached anyway
            else:
                # check if an element still has to be added
                if new_text != "" and new_x0 is not None and new_x1 is not None:
                    self.elements_list.append(
                        BaseElement(x0=new_x0, x1=new_x1, y0=new_y0, y1=new_y1, text=new_text, has_bold=has_bold,
                                    scale_multiplier=self.scale_multiplier))
                    # reset variables
                    new_text = ""
                    new_x0 = None
                    new_x1 = None
                    new_y0 = None
                    new_y1 = None
                self.non_text_elements.append(
                    BaseElement(x0=element.x0, x1=element.x1, y0=element.y0, y1=element.y1, text="",
                                scale_multiplier=self.scale_multiplier))

    def _clean_aligned(self, dict_el, key2_co, alignment_take_min=False, custom_tolerance=None, check_collision=True):

        tolerance_chosen = self.config.TOLERANCE_GEN if custom_tolerance is None else custom_tolerance

        vertical_alignment = True
        if key2_co.startswith("y"):
            vertical_alignment = False

        key_s = key2_co + "_min" if not alignment_take_min else key2_co + "_max"

        xk_pos = list(dict_el.keys())
        len_keys = len(xk_pos)
        to_del = []

        if len_keys <= 1:
            return dict_el

        # make distance matrix
        none_row = [None for _ in range(0, len_keys)]
        dist_matrix = np.array([none_row for _ in range(0, len_keys)])

        distances_dict = {}
        for a in range(0, len_keys):
            for b in range(a + 1, len_keys):
                dist = abs(xk_pos[a] - xk_pos[b])
                if dist <= tolerance_chosen:
                    dist_matrix[a][b] = dist
                    if dist not in distances_dict:
                        distances_dict[dist] = []
                    distances_dict[dist].append((a, b))

        distances_sorted = sorted(list(distances_dict))

        for dist in distances_sorted:
            # consolidate one after the other, updating the none matrix
            to_consolidate = distances_dict[dist]
            for cons in to_consolidate:
                if dist_matrix[cons[0]][cons[1]] is not None:

                    key_taken = xk_pos[cons[0]] if dict_el[xk_pos[cons[0]]]["c"] > dict_el[xk_pos[cons[1]]]["c"] else \
                    xk_pos[cons[1]]
                    key_rejected = xk_pos[cons[0]] if key_taken == xk_pos[cons[1]] else xk_pos[cons[1]]

                    collided = False if not check_collision else has_collision_post_alignment(
                        dict_el[key_taken]['list'], dict_el[key_rejected]['list'], key_taken, vertical_alignment,
                        alignment_take_min)

                    if not collided:
                        to_del.append(key_rejected)
                        # update matrix
                        pos_to_del = xk_pos.index(key_rejected)
                        for kk in range(0, len_keys):
                            dist_matrix[pos_to_del, kk] = None
                            dist_matrix[kk, pos_to_del] = None
                        # update dictionary
                        dict_el[key_taken]["c"] += dict_el[key_rejected]["c"]
                        new_xs = dict_el[key_taken][key_s]
                        switch = dict_el[key_rejected][key_s] < dict_el[key_taken][key_s] if not alignment_take_min else \
                        dict_el[key_rejected][key_s] > dict_el[key_taken][key_s]
                        if switch:
                            new_xs = dict_el[key_rejected][key_s]
                        dict_el[key_taken][key_s] = new_xs
                        dict_el[key_taken]['list'] += dict_el[key_rejected]['list']

        return {k: v for (k, v) in dict_el.items() if k not in to_del}

    def _get_bounding_text_area(self, group: BaseElementGroup, col_elements_exclude, handled=None) -> Tuple[BaseElementGroup, List[BaseElement]]:

        if handled is None:
            handled = [] + group.elements

        row = self.rows[group.row_index]

        for el in row['list']:
            if not el.in_list(group.elements) and not el.in_list(col_elements_exclude) and not el.in_list(handled):
                if abs(el.x0 - group.x1) <= self.config.SPACE_MAX_DISTANCE and el.x1 > group.x1:
                    group.add_element(el)
                    handled.append(el)
                    return self._get_bounding_text_area(group, col_elements_exclude, handled)
                if abs(group.x0 - el.x1) <= self.config.SPACE_MAX_DISTANCE and el.x0 < group.x0:
                    group.add_element(el)
                    handled.append(el)
                    return self._get_bounding_text_area(group, col_elements_exclude, handled)

        return group, handled

    def _find_rows(self):

        dict_bot = {}

        # search for all elements bot aligned
        for element in self.elements_list:
            y0 = math.floor(element.y0)
            y1 = math.ceil(element.y1)
            if y0 not in dict_bot:
                dict_bot[y0] = {"c": 0, "y1_max": None, "list": [], "base_elements": []}
            dict_bot[y0]["c"] += 1
            dict_bot[y0]["y1_max"] = y1 if dict_bot[y0]["y1_max"] is None or dict_bot[y0]["y1_max"] < y1 else \
            dict_bot[y0]["y1_max"]
            dict_bot[y0]['list'].append(element)

        # clean given some tolerance
        dict_bot_cleaned = self._clean_aligned(dict_bot, "y1", True)

        # make list
        rows = []
        for y0_key, dict_n in dict_bot_cleaned.items():
            dict_n['y0_min'] = y0_key
            dict_n['list'] = sorted(dict_n['list'], key=lambda k: k.x0)
            rows.append(dict_n)

        rows = sorted(rows, key=lambda k: -k['y0_min'])

        # assign row index to each element
        for k, row in enumerate(rows):
            for el in row['list']:
                el.row_index = k

        self.rows = rows

        # make base elements for elements in row (bounding text)
        for k, row in enumerate(self.rows):
            base_elements = []
            handled = []
            for el in row['list']:

                if el not in handled:
                    new_el, handled = self._get_bounding_text_area(BaseElementGroup([el]), handled)
                    base_elements.append(new_el)

            row['base_elements'] = base_elements

    def _find_columns_numeric(self, min_row_limit=1):

        dict_numbers = {}

        # search for >= min_row_limit text elements left aligned, >= min_row_limit numbers right aligned
        for element in self.elements_list:
            if is_number_cell(element.text):
                # right alignment
                x0 = math.floor(element.x0)
                x1 = math.ceil(element.x1)
                if x1 not in dict_numbers:
                    dict_numbers[x1] = {"c": 0, "x0_min": None, "list": []}
                dict_numbers[x1]["c"] += 1
                dict_numbers[x1]["x0_min"] = x0 if dict_numbers[x1]["x0_min"] is None or dict_numbers[x1][
                    "x0_min"] > x0 else dict_numbers[x1]["x0_min"]
                dict_numbers[x1]['list'].append(element)

        # clean numbers given some tolerance
        dict_numbers_cleaned = self._clean_aligned(dict_numbers, "x0", False)

        # make list
        number_cols = []
        for x1_key, dict_n in dict_numbers_cleaned.items():
            if dict_n['c'] > min_row_limit:
                dict_n['x1_max'] = x1_key
                number_cols.append(dict_n)

        number_cols = sorted(number_cols, key=lambda k: k['x0_min'])

        self.number_cols_list = number_cols

    def _add_unbroken_areas(self):

        all_col_items = []

        for col_index, col in enumerate(self.number_cols_list):
            all_col_items += col['list']

        for col_index, col in enumerate(self.number_cols_list):

            el_list = []
            # remove elements that are inside a text block
            for kk, el in enumerate(col['list']):
                bounding_text, _ = self._get_bounding_text_area(BaseElementGroup([el]), all_col_items)
                if is_number_cell(bounding_text.text) or len(bounding_text.text) <= self.config.text_max_number_col:
                    el_list.append(el)

            unbroken_areas = []
            # sort items in col by y1 position
            el_list = sorted(el_list, key=lambda x: -x.y1)
            # alignment means at least 2 elements
            alignment_area_start = None
            alignment_count = 0
            element_list = []
            prev_el = None
            for kk, el in enumerate(el_list):

                alignment_count += 1
                element_list.append(el)
                if alignment_count >= 2 and alignment_area_start is None and prev_el is not None:
                    alignment_area_start = prev_el.y1

                if alignment_area_start is not None and prev_el is not None:
                    # check if something is between this element and the one before
                    area_to_check = Area(col['x0_min'], col["x1_max"], el.y1, prev_el.y0)

                    for row in self.rows:
                        break_all = False

                        if area_to_check.y1 + self.config.TOLERANCE_GEN >= row[
                            'y0_min'] >= area_to_check.y0 - self.config.TOLERANCE_GEN or area_to_check.y1 + self.config.TOLERANCE_GEN >= row[
                            'y1_max'] >= area_to_check.y0 - self.config.TOLERANCE_GEN:
                            for el_text_area in row['base_elements']:
                                # check if base element can be used to break the area (can't be used if one of its elements is part of the current column)
                                use_base_el = False if len(
                                    [x for x in el_text_area.elements if x in el_list]) > 0 else True
                                # check if base element collides with current area
                                if use_base_el and area_to_check.collides_with(
                                        el_text_area) and not area_to_check.is_inside(el_text_area):

                                    if alignment_count > 2:
                                        ua = Area(col['x0_min'], col["x1_max"], prev_el.y0, alignment_area_start)
                                        ua.elements = element_list[:-1]
                                        unbroken_areas.append(ua)
                                    alignment_area_start = el.y1
                                    alignment_count = 1
                                    element_list = [el]
                                    break_all = True
                                    break
                        if break_all:
                            break

                    if kk == len(
                            el_list) - 1 and alignment_count >= 2 and prev_el is not None and alignment_area_start is not None:
                        ua = Area(col['x0_min'], col["x1_max"], el.y0, alignment_area_start)
                        ua.elements = element_list
                        unbroken_areas.append(ua)

                prev_el = el

            col['unbroken'] = unbroken_areas

    def _extend_area_obj(self, area_org, upwards, row_index, init=False, row_index_limit=None):

        if init:
            if row_index == 0:
                return
            if row_index == len(self.rows) - 1:
                return
            row_index = row_index - 1 if upwards else row_index + 1
        else:
            if row_index < 0:
                return
            if row_index > len(self.rows) - 1:
                return

        # check that limiting row index is not reached
        if row_index_limit is not None:
            if upwards and row_index <= row_index_limit:
                return
            if not upwards and row_index >= row_index_limit:
                return

        area = copy.deepcopy(area_org)

        if upwards:
            k_1 = "y1_max"
        else:
            k_1 = "y0_min"

        relevant_row = self.rows[row_index]

        # extend y area of area (hypothetically)
        if upwards:
            area.y1 = relevant_row[k_1]
            row_index_next = row_index - 1
        else:
            area.y0 = relevant_row[k_1]
            row_index_next = row_index + 1

        for relevant_el in relevant_row['base_elements']:
            if area.collides_with(relevant_el) and not area.is_inside(relevant_el) and letter_len(relevant_el.text) > 2:
                return

        # extend y area of area (actually)
        if upwards:
            area_org.y1 = area.y1
        else:
            area_org.y0 = area.y0

        self._extend_area_obj(area_org, upwards, row_index_next, False, row_index_limit)

    def _detect_tables(self):

        # delete columns with no unbroken areas
        cols_candidates = [x for x in self.number_cols_list if len(x['unbroken']) > 0]

        # extend unbroken areas for non-numeric elements that fit inside the area
        for kk, col in enumerate(cols_candidates):
            for area_obj_idx, area_obj in enumerate(col['unbroken']):
                # determine boundaries
                # upwards
                row_index_limit = None if area_obj_idx == 0 else max(
                    [x.row_index for x in col['unbroken'][area_obj_idx - 1].elements])
                min_row_index = min([x.row_index for x in area_obj.elements])
                self._extend_area_obj(area_obj, True, min_row_index, True, row_index_limit)

                # downwards
                row_index_limit = None if area_obj_idx == len(col['unbroken']) - 1 else min(
                    [x.row_index for x in col['unbroken'][area_obj_idx + 1].elements])
                max_row_index = max([x.row_index for x in area_obj.elements])
                self._extend_area_obj(area_obj, False, max_row_index, True, row_index_limit)

        # break unbroken columns by blank line
        for k, row in enumerate(self.rows):
            if k > 0:
                distance_to_before = (row['y1_max'] - self.rows[k - 1]["y0_min"]) * -1

                if distance_to_before >= self.config.line_break_distance:
                    break_point_start = self.rows[k - 1]["y0_min"]
                    break_point_end = row['y1_max']
                    for col in cols_candidates:
                        to_add = []
                        for area_obj in col['unbroken']:
                            area_obj, new_obj = area_obj.break_at_horizontal(break_point_start)
                            if new_obj is not None:
                                new_obj, _ = new_obj.break_at_horizontal(break_point_end, "top")
                                to_add.append(new_obj)
                        col['unbroken'] += to_add

        # collect elements inside areas
        all_elements = copy.deepcopy(self.elements_list)
        all_relevant_areas = []
        for col in cols_candidates:
            for area_obj in col['unbroken']:
                num_numbers = 0
                num_text = 0
                must_have_items = area_obj.elements
                area_obj.elements = []
                for k in range(len(all_elements) - 1, -1, -1):
                    if area_obj.is_inside(all_elements[k], 5) or all_elements[k].in_list(must_have_items):
                        area_obj.put_element(all_elements[k])
                        is_number_cell_check = is_number_cell(all_elements[k].text)
                        letter_len_check = letter_len(all_elements[k].text)
                        if is_number_cell_check:
                            num_numbers += 1
                        if not is_number_cell_check and letter_len_check > 0:
                            num_text += 1
                        del (all_elements[k])
                if len(area_obj.elements) > 0 and num_numbers > 0 and num_numbers >= num_text:
                    all_relevant_areas.append(area_obj)

        # save for use in features etc.
        self.numeric_areas = all_relevant_areas

        # all element frames that share some distance on the y-axis form a group (each Area can only be in one group)
        all_relevant_areas = sorted(all_relevant_areas, key=lambda x: -x.y1_el)
        groups = []
        temp_group = None
        for k, ar in enumerate(all_relevant_areas):
            if not ar.in_group:
                if temp_group is None:
                    temp_group = TableGroup([ar], "_el")
                for kk in range(k + 1, len(all_relevant_areas)):
                    ar2 = all_relevant_areas[kk]
                    if not ar2.in_group:
                        if ar2.v_overlap(temp_group, "_el"):
                            temp_group.put_element(ar2, "_el")
                groups.append(temp_group)
                temp_group = None

        # split groups where columns are too far away from each other
        for g in groups:
            g.elements = sorted(g.elements, key=lambda x: x.x0_el)
            break_at = None
            for k, area in enumerate(g.elements):
                if k > 0:
                    if (not self.has_wide_layout() and area.x_distance_to(
                            g.elements[k - 1]) > self.config.separate_columns_distance) or (
                            self.has_wide_layout() and area.x_distance_to(
                            g.elements[k - 1]) > self.config.separate_columns_distance_wide_layout):
                        # separate group
                        break_at = k
                        break
            if break_at is not None:
                new_group_elements = []
                for kk in range(len(g.elements) - 1, break_at - 1, -1):
                    new_group_elements.append(g.elements[kk])
                    g.remove_element(kk, "_el")
                new_group = TableGroup(new_group_elements, "_el")
                groups.append(new_group)

        # combine areas inside group that are horizontally aligned
        for g in groups:
            g.elements = sorted(g.elements, key=lambda x: x.y1_el)
            handled = []
            a = 0
            while a < len(g.elements):
                if a not in handled:
                    for b in range(0, len(g.elements)):
                        if a != b and b not in handled:
                            if g.elements[a].stackable(g.elements[b], "_el"):
                                handled.append(a)
                                handled.append(b)
                                g.elements.append(g.elements[a].combine(g.elements[b]))
                                break
                a += 1

            for index_del in sorted(handled, reverse=True):
                del g.elements[index_del]

        # TABLE AREA IS THE MAX OF ALL ALIGNED UNBROKEN AREAS INSIDE GROUP
        for g in groups:
            x0 = None
            x1 = None
            y0 = None
            y1 = None
            for area in g.elements:
                min_x0 = max(area.x0, g.x0)
                max_x1 = min(area.x1, g.x1)
                min_y0 = max(area.y0, g.y0)
                max_y1 = min(area.y1, g.y1)

                x0 = min_x0 if x0 is None or min_x0 < x0 else x0
                x1 = max_x1 if x1 is None or max_x1 > x1 else x1
                y0 = min_y0 if y0 is None or min_y0 < y0 else y0
                y1 = max_y1 if y1 is None or max_y1 > y1 else y1

            if x0 is not None:
                g.data_area = Area(x0, x1, y0, y1)

        # check that all groups have some line items
        valid_groups = []
        for k, g in enumerate(groups):
            if self.natural_text.has_line_items(g):
               valid_groups.append(k)

        self.groups = [x for k, x in enumerate(groups) if k in valid_groups]

    def _find_columns_text(self, el_list, tolerance=0):

        dict_text = {}

        for element in el_list:
            # left alignment
            x0 = math.floor(element.x0)
            x1 = math.ceil(element.x1)
            if x0 not in dict_text:
                dict_text[x0] = {"c": 0, "x1_max": None, "list": []}
            dict_text[x0]["c"] += 1
            dict_text[x0]["x1_max"] = x1 if dict_text[x0]["x1_max"] is None or dict_text[x0]["x1_max"] < x1 else \
            dict_text[x0]["x1_max"]
            dict_text[x0]['list'].append(element)

        # clean numbers given some tolerance
        dict_text_cleaned = self._clean_aligned(dict_text, "x1", True, tolerance, False)

        # make list
        text_cols = []
        for x0_key, dict_n in dict_text_cleaned.items():
            dict_n['x0_min'] = x0_key
            text_cols.append(dict_n)

        return sorted(text_cols, key=lambda k: k['x0_min'])

    def _detect_line_items(self):

        # make candidates for entire page: left aligned texts
        all_elements = []
        for row in self.rows:
            for el in row['base_elements']:
                # element should not be numeric
                if not is_number_cell(el.text):
                    all_elements.append(el)

        all_candidate_cols = self._find_columns_text(all_elements, self.config.tolerance_columns_li)

        # make areas
        for col in all_candidate_cols:
            col['area'] = Area()
            col['area'].init_with_elements(col['list'])

        # for each group: find best left aligned text -> criteria: distance (the closer the better), contains some text
        candidates_by_group = {}
        for g_index, g in enumerate(self.groups):

            # get rows that actually contain values
            value_rows = []
            for area in g.elements:
                for el in area.elements:
                    if el.row_index not in value_rows:
                        value_rows.append(el.row_index)
            value_rows = sorted(value_rows)

            candidates_by_group[g_index] = []
            # check which candidates are valid for the table group
            for candidate in all_candidate_cols:
                if candidate['area'].x0 < g.x0 and candidate['area'].v_overlap(g):
                    overlapping_elements = [x for x in candidate['list'] if x.row_index in value_rows]
                    if len(overlapping_elements) == 0:
                        continue
                    overlapping_elements_area = Area()
                    overlapping_elements_area.init_with_elements(overlapping_elements)
                    candidates_by_group[g_index].append(
                        {"scoring": {"final_score": 0}, "overlapping_elements_area": overlapping_elements_area,
                         "value_rows": value_rows})

        # criteria for scoring, min letter len of line item of 6
        scoring_weights = {"words": 5, "distance": 1, "completeness": 5, "natural_text_fits": 8}
        chosen_bordering_area = {}
        for g_index, g in enumerate(self.groups):
            # if only one candidate, take that
            if len(candidates_by_group[g_index]) == 1:
                chosen_bordering_area[g_index] = candidates_by_group[g_index][0]
            elif len(candidates_by_group[g_index]) > 1:
                # sort by distance to values
                candidates_by_group[g_index] = list(
                    sorted(candidates_by_group[g_index], key=lambda x: g.x0 - x['overlapping_elements_area'].x0))
                for k, candidate_dict in enumerate(candidates_by_group[g_index]):
                    candidate_dict['scoring']['words'] = len(
                        [x for x in candidate_dict['overlapping_elements_area'].elements if len(words_contained(x.text)) > 0]) / len(candidate_dict['value_rows'])
                    candidate_dict['scoring']['completeness'] = len(
                        [x for x in candidate_dict['overlapping_elements_area'].elements]) / len(
                        candidate_dict['value_rows'])
                    candidate_dict['scoring']['distance'] = 1 if len(candidates_by_group[g_index]) == 1 else 1 - (
                                k / (len(candidates_by_group[g_index]) - 1))
                    candidate_dict['scoring']['natural_text_fits'] = self.natural_text.is_adjacent_percent(g, candidate_dict['overlapping_elements_area'])
                    for key, weight in scoring_weights.items():
                        candidate_dict['scoring']['final_score'] += candidate_dict['scoring'][key] * weight
                scores = [x['scoring']['final_score'] for x in candidates_by_group[g_index]]
                index_taken = scores.index(max(scores))
                chosen_bordering_area[g_index] = candidates_by_group[g_index][index_taken]

            # assign line items
            line_item_helper = {}
            if g_index in chosen_bordering_area:
                for row_index, row in enumerate(self.rows):
                    if row_index in chosen_bordering_area[g_index]['value_rows'] or (
                            (g.y1 > row['y1_max'] > g.y0) or (g.y1 > row['y0_min'] > g.y0)):
                        for el in row['base_elements']:
                            # condition for base element to be added -> doesnt collide with values and is inside the area identified as line item area or on the right of it
                            if not g.is_inside(el) and not g.collides_with(el) and el.x0 < g.x0 and (
                                    chosen_bordering_area[g_index]['overlapping_elements_area'].x0 <= el.x0 or
                                    chosen_bordering_area[g_index]['overlapping_elements_area'].is_inside(el) or
                                    chosen_bordering_area[g_index]['overlapping_elements_area'].h_overlap(el)):
                                # add li object
                                if row_index not in line_item_helper:
                                    line_item_helper[row_index] = LineItem(el)
                                else:
                                    # add to li
                                    line_item_helper[row_index].add_el(el)
                        # check if something was found, otherwise try with actual elements instead of base_elements
                        if row_index not in line_item_helper:
                            for el in row['list']:
                                # condition for base element to be added -> doesnt collide with values and is inside the area identified as line item area or on the right of it
                                if not g.is_inside(el) and not g.collides_with(el) and el.x0 < g.x0 and (
                                        chosen_bordering_area[g_index]['overlapping_elements_area'].x0 <= el.x0 or
                                        chosen_bordering_area[g_index]['overlapping_elements_area'].is_inside(el) or
                                        chosen_bordering_area[g_index]['overlapping_elements_area'].h_overlap(el)):
                                    # add li object
                                    if row_index not in line_item_helper:
                                        line_item_helper[row_index] = LineItem(el)
                                    else:
                                        # add to li
                                        line_item_helper[row_index].add_el(el)

            g.line_items = list(line_item_helper.values())

    def _split_table_if_needed(self, table: ExtractedTable) -> List[ExtractedTable]:

        output_list = []

        split_table = False
        for k, li in enumerate(table.items):
            if k > 0:
                if abs(table.items[k - 1].el.y0 - li.el.y1) > self.config.separate_table_distance:
                    split_table = True
                    table_one = ExtractedTable(table.items[0:k], table.g_index)
                    output_list.append(table_one)
                    # table 2
                    table_two = ExtractedTable(table.items[k:], table.g_index)
                    output_list += self._split_table_if_needed(table_two)
                    break
        if not split_table:
            output_list = [table]

        return output_list

    def extract_tables(self, min_rows: int = 1, min_cols: int = 1) -> List[ExtractedTable]:

        self._find_rows()
        self._find_columns_numeric()
        self._add_unbroken_areas()
        self._detect_tables()
        self._detect_line_items()

        tables = []

        for g_index, g in enumerate(self.groups):

            if len(g.elements) < min_cols:
                continue

            # sort elements
            g.elements = sorted(g.elements, key=lambda x: x.x1_el)

            value_grid = {row_index: [None for _ in range(0, len(g.elements))] for row_index in g.elements_by_row().keys()}

            for k, area in enumerate(g.elements):

                area.elements = sorted(area.elements, key=lambda x: (x.x0, -x.y1))

                for el in area.elements:

                    # find bounding el in row
                    bounding_el = None
                    for bounding_el_row in self.rows[el.row_index]['base_elements']:
                        bounding_identical = bounding_el_row.is_identical(el)
                        if bounding_identical or el.in_list(bounding_el_row.elements):
                            bounding_el = bounding_el_row
                            break

                    if bounding_el is not None:

                        # if bounding el contains more than one number cell, take the original element instead of bounding
                        if len([None for x in bounding_el.elements if is_number_cell(x.text)]) > 1:
                            bounding_el = el

                        # check that element is really in grid
                        if bounding_el.row_index not in value_grid:
                            continue

                        if value_grid[bounding_el.row_index][k] is None:
                            value_grid[bounding_el.row_index][k] = bounding_el
                        else:
                            if value_grid[bounding_el.row_index][k].text == bounding_el.text:
                                continue
                            else:
                                # merge elements
                                value_grid[bounding_el.row_index][k] = copy.deepcopy(value_grid[bounding_el.row_index][k])
                                value_grid[bounding_el.row_index][k].merge(bounding_el)

            # create table elements
            final_table = ExtractedTable(g.line_items, g_index)

            # fill empty line items
            final_table.fill_empty_li(value_grid)

            # separate table if li's too far away from each other
            separate_final_tables = self._split_table_if_needed(final_table)

            # filter out empty columns
            separate_final_tables = filter_out_empty_columns(separate_final_tables, min_cols)

            tables += separate_final_tables

        # sort
        tables = sorted(tables, key=lambda x: -x.table_area.y1)

        # discard tables with not enough rows
        to_del = []
        for k, table in enumerate(tables):
            if len(table.items) < min_rows:
                to_del.append(k)

        for k in range(len(tables) - 1, -1, -1):
            if k in to_del:
                del (tables[k])

        return tables

    def extract_text_and_tables(self, min_rows: int = 1, min_cols: int = 1, **kwargs) -> List[ExtractedPdfElement]:

        if min_cols < 1 or min_rows < 1:
            raise Exception("a table needs at least one column and one row")

        # launch table recognition
        tables = self.extract_tables(min_rows, min_cols)

        all_elements = []
        all_extracted_elements: List[ExtractedPdfElement] = []

        for row in self.rows:
            for base_el in row['base_elements']:
                # check if element collides with a table
                in_table = False
                for t in tables:
                    if t.table_area.collides_with(base_el) or t.table_area.is_inside(base_el):
                        in_table = True
                        # add table now also to all elements
                        if t not in all_elements:
                            table_inserted = False
                            # CHECK FIRST IF TABLE WAS SPLIT, IF SO, SORT AFTER THE FIRST SPLIT TABLE
                            all_tables = [(k, x.g_index) for k, x in enumerate(all_elements) if
                                          isinstance(x, ExtractedTable)]
                            if t.g_index in [x[1] for x in all_tables]:
                                # find last occurrence of g_index
                                idx = None
                                for kk in range(len(all_tables) - 1, -1, -1):
                                    if all_tables[kk][1] == t.g_index:
                                        idx = all_tables[kk][0]
                                        break
                                if idx is not None:
                                    table_inserted = True
                                    all_elements.insert(idx + 1, t)
                                    all_extracted_elements.insert(idx + 1, t)
                            # add at the end if not inserted through previous g_index
                            if not table_inserted:
                                all_elements.append(t)
                                all_extracted_elements.append(t)
                        # check if element is inside line item area
                        if t.line_item_area.is_inside(base_el) or t.line_item_area.overlap_percent(base_el) > 0.8:
                            # check if element was used
                            if not t.line_item_area.contains(base_el):
                                t.add_to_items(base_el)
                        # check if element is inside value area
                        elif t.total_value_area.is_inside(base_el):
                            # check if element was placed
                            if base_el not in t.total_value_area.elements:
                                # try to place element
                                for col_idx, area in enumerate(t.value_areas):
                                    if area.collides_with(base_el) or area.is_inside(base_el):
                                        if base_el not in area.elements:
                                            # element is not placed yet, check if there is space
                                            if base_el.row_index not in area.all_row_indices:
                                                area.put_element(base_el)
                                                t.add_value(base_el, col_idx)
                                                break
                        else:
                            in_table = False
                        break
                # add text element
                if not in_table:
                    all_extracted_elements.append(ExtractedPdfElement(base_el.x0, base_el.x1, base_el.y0, base_el.y1, base_el))

        return all_extracted_elements
